part of firebase_model_notifier;

abstract class FirebaseFunctionsCollectionModel<T> extends ValueModel<List<T>>
    with ListModelMixin<T>
    implements StoredModel<List<T>> {
  FirebaseFunctionsCollectionModel(this.endpoint, [List<T>? value])
      : super(value ?? []);

  @protected
  FirebaseFunctions get functions {
    return FirebaseFunctions.instance;
  }

  @override
  bool get notifyOnChangeList => false;

  @override
  bool get notifyOnChangeValue => true;

  final String endpoint;

  List<T> fromCollection(List<Object> list);
  List<Object> toCollection(List<T> list);

  @protected
  @mustCallSuper
  Future<void> onLoad() async {}

  @protected
  @mustCallSuper
  Future<void> onSave() async {}

  @protected
  @mustCallSuper
  Future<void> onDidLoad() async {}

  @protected
  @mustCallSuper
  Future<void> onDidSave() async {}

  @protected
  @mustCallSuper
  @protected
  @mustCallSuper
  void onCatchResponse(HttpsCallableResult<String> response) {}

  @protected
  List<Object> fromResponse(String json) => jsonDecodeAsList(json);

  @protected
  dynamic toRequest(List<Object> list) => jsonEncode(list);

  @protected
  @mustCallSuper
  List<Object> filterOnLoad(List<Object> loaded) => loaded;

  @protected
  @mustCallSuper
  List<Object> filterOnSave(List<Object> save) => save;

  @protected
  T createDocument();

  T create() => createDocument();

  @override
  Future<List<T>> load() async {
    await onLoad();

    final res =
        await functions.httpsCallable(endpoint.split("/").last).call<String>();
    onCatchResponse(res);
    final data = fromCollection(filterOnLoad(fromResponse(res.data)));
    addAll(data);
    streamController.sink.add(value);
    notifyListeners();
    await onDidLoad();
    return this;
  }

  @override
  Future<List<T>> save() async {
    await onSave();
    final res =
        await functions.httpsCallable(endpoint.split("/").last).call<String>(
              toRequest(filterOnSave(toCollection(this))),
            );
    onCatchResponse(res);
    streamController.sink.add(value);
    notifyListeners();
    await onDidSave();
    return this;
  }
}
