part of firebase_model_notifier;

abstract class FirebaseFunctionsDocumentModel<T> extends DocumentModel<T>
    implements StoredModel<T> {
  FirebaseFunctionsDocumentModel(this.endpoint, T initialValue)
      : super(initialValue);

  @override
  @protected
  @mustCallSuper
  void initState() {
    super.initState();
    value ??= initialValue;
  }

  @protected
  FirebaseFunctions get functions {
    return FirebaseFunctions.instance;
  }

  @protected
  T get initialValue;

  final String endpoint;

  @protected
  @mustCallSuper
  Future<void> onLoad() async {}

  @protected
  @mustCallSuper
  Future<void> onSave() async {}

  @protected
  @mustCallSuper
  Future<void> onDidLoad() async {}

  @protected
  @mustCallSuper
  Future<void> onDidSave() async {}

  @protected
  @mustCallSuper
  void onCatchResponse(HttpsCallableResult<String> response) {}

  @protected
  Map<String, dynamic> fromResponse(String json) => jsonDecodeAsMap(json);

  @protected
  dynamic toRequest(Map<String, dynamic> map) => jsonEncode(map);

  @protected
  @mustCallSuper
  Map<String, dynamic> filterOnLoad(Map<String, dynamic> loaded) => loaded;

  @protected
  @mustCallSuper
  Map<String, dynamic> filterOnSave(Map<String, dynamic> save) => save;

  @override
  Future<T> load() async {
    await onLoad();

    final res =
        await functions.httpsCallable(endpoint.split("/").last).call<String>();
    onCatchResponse(res);
    value = fromMap(filterOnLoad(fromResponse(res.data)));
    streamController.sink.add(value);
    notifyListeners();
    await onDidLoad();
    return value;
  }

  @override
  Future<T> save() async {
    await onSave();
    final res =
        await functions.httpsCallable(endpoint.split("/").last).call<String>(
              toRequest(filterOnSave(toMap(value))),
            );
    onCatchResponse(res);
    streamController.sink.add(value);
    notifyListeners();
    await onDidSave();
    return value;
  }

  @override
  // ignore: avoid_equals_and_hash_code_on_mutable_classes
  bool operator ==(Object other) => hashCode == other.hashCode;

  @override
  // ignore: avoid_equals_and_hash_code_on_mutable_classes
  int get hashCode => super.hashCode ^ endpoint.hashCode;
}
