part of firebase_model_notifier;

abstract class FirebaseFunctionsModel<T> extends ValueModel<T>
    implements StoredModel<T> {
  FirebaseFunctionsModel(this.endpoint, T initialValue) : super(initialValue);

  @protected
  FirebaseFunctions get functions {
    return FirebaseFunctions.instance;
  }

  final String endpoint;

  @protected
  @mustCallSuper
  Future<void> onLoad() async {}

  @protected
  @mustCallSuper
  Future<void> onSave() async {}

  @protected
  @mustCallSuper
  Future<void> onDidLoad() async {}

  @protected
  @mustCallSuper
  Future<void> onDidSave() async {}

  @protected
  @mustCallSuper
  void onCatchResponse(HttpsCallableResult<String> response) {}

  @protected
  T fromResponse(String data);

  @protected
  dynamic toRequest(T data);

  @protected
  @mustCallSuper
  T filterOnLoad(T loaded) => loaded;

  @protected
  @mustCallSuper
  T filterOnSave(T save) => save;

  @override
  Future<T> load() async {
    await onLoad();

    final res =
        await functions.httpsCallable(endpoint.split("/").last).call<String>();
    onCatchResponse(res);
    value = filterOnLoad(fromResponse(res.data));
    streamController.sink.add(value);
    notifyListeners();
    await onDidLoad();
    return value;
  }

  @override
  Future<T> save() async {
    await onSave();
    final res =
        await functions.httpsCallable(endpoint.split("/").last).call<String>(
              toRequest(filterOnSave(value)),
            );
    onCatchResponse(res);
    streamController.sink.add(value);
    notifyListeners();
    await onDidSave();
    return value;
  }

  @override
  // ignore: avoid_equals_and_hash_code_on_mutable_classes
  bool operator ==(Object other) => hashCode == other.hashCode;

  @override
  // ignore: avoid_equals_and_hash_code_on_mutable_classes
  int get hashCode => super.hashCode ^ endpoint.hashCode;
}
